# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

use Test::More;

BEGIN { use_ok('Search::Elasticsearch::Async') }

my ( $e, $p, $t );

ok $e = Search::Elasticsearch::Async->new(), "new client";
ok $e->does('Search::Elasticsearch::Role::Client::Direct'),
    "client does Search::Elasticsearch::Role::Client::Direct";

isa_ok $t = $e->transport, 'Search::Elasticsearch::Transport::Async',
    "transport";
isa_ok $p = $t->cxn_pool, 'Search::Elasticsearch::CxnPool::Async::Static',
    "cxn_pool";
isa_ok $p->cxn_factory, 'Search::Elasticsearch::Cxn::Factory', "cxn_factory";
isa_ok $e->logger, 'Search::Elasticsearch::Logger::LogAny', "logger";

done_testing;

